local crime_type_names = {
	'Violation of Production Order',
	'Violation of Export Prohibition',
	'Violation of Job Order',
	'Conspiracy to Slow Labor',
	'Murder', --of ...
	'Disorderly Conduct',
	'Building Destruction',
	'Vandalism',
	'Theft',
	'Robbery',
	'Blood-drinking'
}

--luacheck: in=
function justice_get_data()	
    if not have_noble('SHERIFF') and not have_noble('CAPTAIN_OF_THE_GUARD') then
        return { false }
    end

    return execute_with_status_page(status_pages.Justice, function(ws)
    	local ws = ws --as:df.viewscreen_justicest
    	
		local function process_crimes(ret, cases)
			for i,v in ipairs(ws.cases) do
				local victim = df.unit.find(v.victim)
				local victim_name = v.victim ~= -1 and unit_fulltitle(victim) or '' --todo: mp.NIL here and check in the app

				local convict = df.unit.find(v.convicted)
				local convict_name = v.convicted ~= -1 and unit_fulltitle(convict) or mp.NIL

				local accused = {}
				local accused_cnt = 0
				for j,w in ipairs(v.reports) do
					--[[local witness = df.unit.find(w.witness)
					local witness_name = unit_fulltitle(witness)

					local accused = df.unit.find(w.accuses)
					local accused_name = unit_fulltitle(accused)]]

					if not accused[w.accuses] then
						accused[w.accuses] = true
						accused_cnt = accused_cnt + 1
					end

					--table.insert(witnesses, { witness_name, witness.id, accused_name, accused.id })
				end

				table.insert(ret,
					{ crime_type_names[v.mode+1], v.id, v.mode,
					  victim_name, v.victim,
					  #v.reports, accused_cnt,
					  convict_name, v.convicted,
					  v.flags.needs_trial })
			end
		end

		local recent_cases = {}
		process_crimes(recent_cases)

		local cold_cases = {}
		gui.simulateInput(ws, K'CHANGETAB')
		process_crimes(cold_cases)

		local convicts = {}
		for i,v in ipairs(ws.convicts) do
			local fullname = unit_fulltitle(v)

			local punishment = mp.NIL
			for j,w in ipairs(df.global.ui.punishments) do
				if v.id == w.criminal.id then
					punishment = { w.beating, w.hammer_strikes, math.ceil((w.prison_counter+1)/TU_PER_DAY*10) } --todo: why *10 ?
				end
			end

			table.insert(convicts, { fullname, v.id, punishment, C_unit_dead(v) })
		end

		return { true, recent_cases, cold_cases, convicts, ws.jails_needed, ws.jails_present }
	end)
end

--luacheck: in=number
function justice_get_crime_details(crimeid)
	local v = df.crime.find(crimeid)
	if not v then
		error('no crime '..tostring(crimeid))
	end

	local victim = df.unit.find(v.victim)
	local victim_name = v.victim ~= -1 and unit_fulltitle(victim) or '' --todo: mp.NIL here and check in the app

	local convict = df.unit.find(v.convicted)
	local convict_name = v.convicted ~= -1 and unit_fulltitle(convict) or mp.NIL

	local witnesses = {}
	for j,w in ipairs(v.reports) do
		local witness = df.unit.find(w.witness)
		local witness_name = w.witness ~= -1 and unit_fulltitle(witness) or '' --todo: mp.NIL here and check in the app

		local accused = df.unit.find(w.accuses)
		local accused_name = w.accuses ~= -1 and unit_fulltitle(accused) or mp.NIL

		local event_str = format_date(w.event_year, w.event_time)
		local report_str = format_date(w.report_year, w.report_time)

		local found_body = C_crime_report_found_body(w)

		table.insert(witnesses, { witness_name, w.witness, accused_name, w.accuses, event_str, report_str, found_body })
	end

	return { crime_type_names[v.mode+1], v.id, v.mode,
			 victim_name, v.victim,
			 convict_name, v.convicted, convict and C_unit_dead(convict) or false,
			 witnesses, v.flags.needs_trial }
end

--luacheck: in=number
function justice_get_convict_info(unitid)
	local unit = df.unit.find(unitid)
	if not unit then
		error('no unit '..tostring(unitid))
	end

	local officer = nil
	local officer_name = mp.NIL

	local punishment = mp.NIL
	for j,w in ipairs(df.global.ui.punishments) do
		if w.criminal.id == unitid then
			punishment = { w.beating, w.hammer_strikes, math.ceil((w.prison_counter+1)/TU_PER_DAY*10) } --todo: why *10 ?

			officer = w.officer
			officer_name = officer and unit_fulltitle(officer) or mp.NIL

			break
		end
	end

	local crimes = {}
	for i,v in ipairs(df.global.world.crimes.all) do
		if v.convicted == unitid then
			local victim = df.unit.find(v.victim)
			local victim_name = victim and unit_fulltitle(victim) or mp.NIL

			table.insert(crimes, { crime_type_names[v.mode+1], v.id, v.mode, victim_name, victim and victim.id or -1 })
		end
	end

	return { unit_fulltitle(unit), unit.id, C_unit_dead(unit), punishment, crimes, officer_name, officer and officer.id or -1 }
end

local function focus_crime(ws, crimeid)
	local idx = -1
	for i,v in ipairs(ws.cases) do
		if v.id == crimeid then
			idx = i
			break
		end
	end

	if idx == -1 then
		gui.simulateInput(ws, K'CHANGETAB')
		for i,v in ipairs(ws.cases) do
			if v.id == crimeid then
				idx = i
				break
			end
		end
	end

	if idx == -1 then
		return nil
	end

	ws.sel_idx_current = idx

	return ws.cases[idx]
end

--luacheck: in=number,bool,bool
function justice_get_convict_choices(crimeid, show_innocent, show_dead)
	show_innocent = istrue(show_innocent)
	show_dead = istrue(show_dead)

    return execute_with_status_page(status_pages.Justice, function(ws)
    	local ws = ws --as:df.viewscreen_justicest
    	local crime = focus_crime(ws, crimeid)

    	if not crime or crime.convicted ~= -1 or not crime.flags.needs_trial then
    		error('no crime or convicted or no trial '..tostring(crimeid))
    	end

		gui.simulateInput(ws, K'SELECT')	

		if ws.cur_column ~= 2 then
			error('can not switch to choices list')
		end

		local ret = {}
		for i,unit in ipairs(ws.convict_choices) do
			if show_dead or not C_unit_dead(unit) then
				local name = unit_fulltitle(unit)

				local wcnt = 0
				for j,w in ipairs(crime.reports) do
					if w.accuses == unit.id then
						wcnt = wcnt + 1
					end
				end

				if show_innocent or wcnt > 0 then
					table.insert(ret, { name, unit.id, C_unit_dead(unit), wcnt })
				end
			end
		end

		return ret
	end)
end

--luacheck: in=number,number
function justice_convict(crimeid, unitid)
    return execute_with_status_page(status_pages.Justice, function(ws)
    	local ws = ws --as:df.viewscreen_justicest
    	local crime = focus_crime(ws, crimeid)

    	if not crime or crime.convicted ~= -1 or not crime.flags.needs_trial then
    		error('no crime or convicted or no trial '..tostring(crimeid))
    	end

		gui.simulateInput(ws, K'SELECT')	

		if ws.cur_column ~= 2 then
			error('can not switch to choices list')
		end

		for i,unit in ipairs(ws.convict_choices) do
			if unit.id == unitid then
				ws.cursor_right = i
				gui.simulateInput(ws, K'SELECT')

				return true
			end
		end

		error('can not find unit '..tostring(unitid))
	end)	
end

--print(pcall(function() return json:encode(justice_get_data()) end))
--print(pcall(function() return json:encode(justice_get_convict_info(2520)) end))
--print(pcall(function() return json:encode(justice_get_crime_details(1)) end))
--print(pcall(function() return json:encode(justice_get_convict_choices(33)) end))
